#include "stdafx.h"
#include "ControlAutoLayout.h"


ControlAutoLayout::ControlAutoLayout(void)
{
    wnd_dialog_ = NULL;

    is_auto_layout_ = true;

    rect_client_window_original_ = CRect(0,0,0,0);
    rect_client_window_now_ = CRect(0,0,0,0);

    horizontal_ratio_ = 0.0;
    vertical_ratio_ = 0.0;

    map_control_original_rect_.clear();
}


ControlAutoLayout::~ControlAutoLayout(void)
{

}



void ControlAutoLayout::SetControlAutoChangeMode(int ctrl_id, int left_change_mode, 
        int right_change_mode, int top_change_mode, int bottom_change_mode) 
{
    if (!is_auto_layout_)
    {
        return;
    }

    if (!wnd_dialog_)
    {
        return;
    }

    CWnd* control_item = wnd_dialog_->GetDlgItem(ctrl_id);
    if (!control_item)
    {
        return;
    }

    int left_coord;
    int right_coord;
    int top_coord;
    int bottom_coord;

    InitControlRectInMap(control_item);
    CRect control_item_rect = GetControlRect(control_item);

    GetRectCoord(control_item_rect, 
        left_coord, right_coord, top_coord, bottom_coord);

    CalcHorizontalCoord(left_change_mode, right_change_mode, 
        left_coord, right_coord);

    CalcVerticalCoord(top_change_mode, bottom_change_mode,
        top_coord, bottom_coord);


    SetRectByCoord(control_item_rect, 
        left_coord, right_coord, top_coord, bottom_coord);

    control_item->MoveWindow(control_item_rect);
}  


void ControlAutoLayout::SetDialogHandle(CWnd *wnd_dialog)
{
    wnd_dialog_ = wnd_dialog;
}

void ControlAutoLayout::SetAutoLayoutEnable(bool is_auto_layout)
{
    is_auto_layout_ = is_auto_layout;
}

void ControlAutoLayout::UpateControlOriginalRect(CWnd* control_item, CRect rect)
{
    if (NULL == control_item)
    {
        return;
    }

    map_control_original_rect_[control_item] = rect;
}


void ControlAutoLayout::InitLayout(int window_width, int window_height)
{
    //set the prev rect window first
    SetClientWindowRectOriginal(window_width, window_height);

    SetClientWindowRectNow(window_width, window_height);

    CalcHorizontalRatio();

    CalcVerticalRatio();
}


void ControlAutoLayout::SetClientWindowRectOriginal(
    int window_width, int window_height)
{
    //the first time to call it, set the current window to original
    if (0 == rect_client_window_original_.right
        && 0 == rect_client_window_original_.bottom)
    {
        rect_client_window_original_.left = 0;
        rect_client_window_original_.top = 0;
        rect_client_window_original_.right = window_width;
        rect_client_window_original_.bottom = window_height;
    }
}


void ControlAutoLayout::SetClientWindowRectNow(
    int window_width, int window_height)
{
    rect_client_window_now_.left = 0;
    rect_client_window_now_.top = 0;
    rect_client_window_now_.right = window_width;
    rect_client_window_now_.bottom = window_height;
}


void ControlAutoLayout::CalcHorizontalRatio()
{
    double width_now = (double)rect_client_window_now_.Width();
    double width_prev = (double)rect_client_window_original_.Width();
    if (IsEqual(width_prev, 0.0))
    {
        horizontal_ratio_ = 0.0;
        return;
    }

    horizontal_ratio_ = width_now/width_prev;
    return; 
}

void ControlAutoLayout::CalcVerticalRatio()
{
    double height_now = (double)rect_client_window_now_.Height();
    double height_prev = (double)rect_client_window_original_.Height();
    if (IsEqual(height_prev, 0.0))
    {
        vertical_ratio_ = 0.0;
        return;
    }

    vertical_ratio_ = height_now/height_prev;
    return; 

}


void ControlAutoLayout::CalcHorizontalCoord(int left_change_mode, 
                                         int right_change_mode, int &left_coord, int &right_coord)
{
    int horizontal_change_mode = 
        GetHorizontalChangeMode(left_change_mode, right_change_mode);

    switch (horizontal_change_mode)
    {
    case LEFT_PADDING_RIGHT_PADDING:
        CalcLeftPadRightPad(left_coord, right_coord);
        break;
    case LEFT_PADDING_RIGHT_RATIO:
        CalcLeftPadRightRatio(left_coord, right_coord);
        break;
    case LEFT_PADDING_RIGHT_FIXED_WIDTH:
        //do nothing
        break;
    case LEFT_RATIO_RIGHT_PADDING:
        CalcLeftRatioRightPad(left_coord, right_coord);
        break;
    case LEFT_RATIO_RIGHT_RATIO:
        CalcLeftRatioRightRatio(left_coord, right_coord);
        break;
    case LEFT_RATIO_RIGHT_FIXED_WIDTH:
        CalcLeftRatioRightFixedWidth(left_coord, right_coord);
        break;
    case LEFT_FIXED_WIDTH_RIGHT_PADDING:
        CalcLeftFixedWidthRightPad(left_coord, right_coord);
        break;
    case LEFT_FIXED_WIDTH_RIGHT_RATIO:
        CalcLeftFixedWidthRightRatio(left_coord, right_coord);
        break;
    case LEFT_FIXED_WIDTH_RIGHT_FIXED_WIDTH:
        //do nothing
        break;
    default:
        break;
    }
}

void ControlAutoLayout::CalcVerticalCoord(int top_change_mode, int bottom_change_mode,
                                       int &top_coord, int &bottom_coord)
{
    int vertical_change_mode = 
        GetVerticalChangeMode(top_change_mode, bottom_change_mode);

    switch (vertical_change_mode)
    {
    case TOP_PADDING_BOTTOM_PADDING:
        CalcTopPadBottomPad(top_coord, bottom_coord);
        break;
    case TOP_PADDING_BOTTOM_RATIO:
        CalcTopPadBottomRatio(top_coord, bottom_coord);
        break;
    case TOP_PADDING_BOTTOM_FIXED_HEIGHT:
        //do nothing
        break;
    case TOP_RATIO_BOTTOM_PADDING:
        CalcTopRatioBottomPad(top_coord, bottom_coord);
        break;
    case TOP_RATIO_BOTTOM_RATIO:
        CalcTopRatioBottomRatio(top_coord, bottom_coord);
        break;
    case TOP_RATIO_BOTTOM_FIXED_HEIGHT:
        CalcTopRatioBottomFixedHeight(top_coord, bottom_coord);
        break;
    case TOP_FIXED_HEIGHT_BOTTOM_PADDING:
        CalcTopFixedHeightBottomPad(top_coord, bottom_coord);
        break;
    case TOP_FIXED_HEIGHT_BOTTOM_RATIO:
        CalcTopFixedHeightBottomRatio(top_coord, bottom_coord);
        break;
    case TOP_FIXED_HEIGHT_BOTTOM_FIXED_HEIGHT:
        //do nothing
        break;
    default:
        break;
    }

}


//////////////////////////////////////////////////////////////////////////
void ControlAutoLayout::CalcLeftPadRightPad(int &left_coord, int &right_coord)
{
    //left coordinate does not need change
    //left_coord = left_coord;

    int right_pad_width_orginal = rect_client_window_original_.Width() - right_coord;
    right_coord = rect_client_window_now_.Width() - right_pad_width_orginal;
}

void ControlAutoLayout::CalcLeftPadRightRatio(int &left_coord, int &right_coord)
{
    //left coordinate does not need change
    //left_coord = left_coord;

    int right_pad_width_orignal = rect_client_window_original_.Width() - right_coord;
    int right_pad_width_new = (int)(right_pad_width_orignal * horizontal_ratio_);
    right_coord = rect_client_window_now_.Width() - right_pad_width_new;
}

void ControlAutoLayout::CalcLeftRatioRightPad(int &left_coord, int &right_coord)
{
    left_coord = (int)(left_coord * horizontal_ratio_);

    int right_pad_width_orginal = rect_client_window_original_.Width() - right_coord;
    right_coord = rect_client_window_now_.Width() - right_pad_width_orginal;
}

void ControlAutoLayout::CalcLeftRatioRightRatio(int &left_coord, int &right_coord)
{
    left_coord = (int)(left_coord * horizontal_ratio_);

    int right_pad_width_orignal = rect_client_window_original_.Width() - right_coord;
    int right_pad_width_new = (int)(right_pad_width_orignal * horizontal_ratio_);
    right_coord = rect_client_window_now_.Width() - right_pad_width_new;
}


void ControlAutoLayout::CalcLeftRatioRightFixedWidth(int &left_coord, int &right_coord)
{
    int width_original = right_coord - left_coord;

    left_coord = (int)(left_coord * horizontal_ratio_);

    right_coord = left_coord + width_original;
}


void ControlAutoLayout::CalcLeftFixedWidthRightPad(int &left_coord, int &right_coord)
{
    int width_original = right_coord - left_coord;
    int right_pad_width_original = rect_client_window_original_.Width() - right_coord;

    right_coord = rect_client_window_now_.Width() - right_pad_width_original;

    left_coord = right_coord - width_original;
}


void ControlAutoLayout::CalcLeftFixedWidthRightRatio(int &left_coord, int &right_coord)
{
    int width_original = right_coord - left_coord;
    int right_pad_width_original = rect_client_window_original_.Width() - right_coord;
    int right_pad_width_new = (int)(right_pad_width_original * horizontal_ratio_);

    right_coord = rect_client_window_now_.Width() - right_pad_width_new;

    left_coord = right_coord - width_original;
}

//////////////////////////////////////////////////////////////////////////

void ControlAutoLayout::CalcTopPadBottomPad(int &top_coord, int &bottom_coord)
{
    //
    top_coord = top_coord;

    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;
    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_original;
}


void ControlAutoLayout::CalcTopPadBottomRatio(int &top_coord, int &bottom_coord)
{
    //
    top_coord = top_coord;

    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;
    int bottom_pad_height_new = (int)(bottom_pad_height_original * vertical_ratio_);

    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_new;
}


void ControlAutoLayout::CalcTopRatioBottomPad(int &top_coord, int &bottom_coord)
{
    top_coord = (int)(top_coord * vertical_ratio_);

    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;
    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_original;
}


void ControlAutoLayout::CalcTopRatioBottomRatio(int &top_coord, int &bottom_coord)
{
    top_coord = (int)(top_coord * vertical_ratio_);

    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;
    int bottom_pad_height_new = (int)(bottom_pad_height_original * vertical_ratio_);
    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_new;
}

void ControlAutoLayout::CalcTopRatioBottomFixedHeight(int &top_coord, int &bottom_coord)
{
    int height_original = rect_client_window_original_.Height() - top_coord;

    top_coord = (int)(top_coord * vertical_ratio_);

    bottom_coord = rect_client_window_now_.Height() - height_original;
}


void ControlAutoLayout::CalcTopFixedHeightBottomPad(int &top_coord, int &bottom_coord)
{
    int height_original = bottom_coord - top_coord;
    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;

    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_original;

    top_coord = bottom_coord - height_original;
}


void ControlAutoLayout::CalcTopFixedHeightBottomRatio(int &top_coord, int &bottom_coord)
{
    int height_original = bottom_coord - top_coord;
    int bottom_pad_height_original = rect_client_window_original_.Height() - bottom_coord;
    int bottom_pad_height_new = (int)(bottom_pad_height_original * vertical_ratio_);

    bottom_coord = rect_client_window_now_.Height() - bottom_pad_height_new;

    top_coord = bottom_coord - height_original;
}


//////////////////////////////////////////////////////////////////////////
void ControlAutoLayout::InitControlRectInMap(CWnd* control_item)
{
    CRect rect_control_item;
    if (NULL == control_item || NULL == wnd_dialog_)
    {
        return;
    }

    if (!IsExistControlRect(control_item))
    {
        control_item->GetWindowRect(&rect_control_item);
        wnd_dialog_->ScreenToClient(&rect_control_item);
        map_control_original_rect_[control_item] = rect_control_item;
    }
}

CRect ControlAutoLayout::GetControlRect(CWnd* control_item)
{
    CRect rect_control_item;
    if (NULL == control_item || NULL == wnd_dialog_)
    {
        return rect_control_item;
    }

    if (IsExistControlRect(control_item))
    {
        rect_control_item = map_control_original_rect_[control_item];
    }
    else
    {
        control_item->GetWindowRect(&rect_control_item);
        wnd_dialog_->ScreenToClient(&rect_control_item);
    }

    return rect_control_item;
}


int ControlAutoLayout::GetHorizontalChangeMode(
    int left_change_mode, int right_change_mode)
{
    //combine left mode and right mode
    // top_change_mode = 0, bottom_change_mode = 1
    //  =>   vertical_change_mode = 0x0001
    int horizontal_change_mode = (left_change_mode << 8) + right_change_mode;

    return horizontal_change_mode;
}


int ControlAutoLayout::GetVerticalChangeMode(int top_change_mode, int bottom_change_mode)
{
    //combine top mode and bottom mode
    // top_change_mode = 1, bottom_change_mode = 2 
    //  =>   vertical_change_mode = 0x0102
    int vertical_change_mode = (top_change_mode << 8) + bottom_change_mode;
    return vertical_change_mode;
}


void ControlAutoLayout::GetRectCoord(CRect rect, 
                                  int &left_coord, int &right_coord, int &top_coord, int &bottom_coord)
{
    left_coord = rect.left;
    right_coord = rect.right;
    top_coord = rect.top;
    bottom_coord = rect.bottom;
}

void ControlAutoLayout::SetRectByCoord(CRect &rect, 
                                    int left_coord, int right_coord, int top_coord, int bottom_coord)
{
    rect.left = left_coord;
    rect.top = top_coord;


    if (kcontrol_width_min > right_coord - left_coord)
    {
        rect.right = left_coord + kcontrol_width_min;
    }
    else
    {
        rect.right = right_coord;
    }


    if (kcontrol_height_min > bottom_coord - top_coord)
    {
        rect.bottom = top_coord + kcontrol_height_min;
    }
    else
    {
        rect.bottom = bottom_coord; 
    }
}


bool ControlAutoLayout::IsExistControlRect(CWnd* control_item)
{
    std::map<CWnd*,CRect>::iterator itr;
    itr = map_control_original_rect_.find(control_item);
    if (itr != map_control_original_rect_.end())
    {
        return true;
    }

    return false;
}


bool ControlAutoLayout::IsEqual(double x, double y)
{
    double differ = std::abs(x - y);

    if (kdouble_precision >= differ)
    {
        return true;
    }

    return false;
}